@@ -3,6 +3,10 @@ import io
 from ar.utils import (
     arbinenc, arintenc, arbindec, arintdec, b64enc_if_not_str, b64enc, b64dec
 )
+from ar.utils.deep_hash import deep_hash
+from .chunk import Chunk
+from .transaction import Transaction
+from . import FORK_2_4
 
 # note: there are actually three different block formats in chain history,
 # according to https://docs.arweave.org/developers/server/http-api#block-format
@@ -64,7 +68,7 @@ class Block:
         if attr.endswith('_raw'):
             return b64dec(getattr(self, attr[:-4]))
         else:
-            return setup().__getattr__(attr)
+            return super().__getattr__(attr)
 
     @property
     def usd_to_ar_rate(self):
@@ -87,7 +91,7 @@ class Block:
         # pull out poa keys for now
         kwparams.update({
             'poa_' + key : value
-            for key, value in kwparams.pop('poa').items()
+            for key, value in kwparams.pop('poa', {}).items()
         })
 
         # convert integer strings to integers
@@ -148,8 +152,7 @@ class Block:
 
         # either 32-byte txids or complete txs
         txs_count = int.from_bytes(stream.read(2), 'big')
-        txs_raw = [arbindec(stream, 24) for idx in range(txs_count)]
-        txs = [b64enc(tx) for tx in txs_raw[::-1]]
+        txs = [Transaction.fromstream(stream) for idx in range(txs_count)][::-1]
 
         return cls(indep_hash = indep_hash_raw, prev_block = prev_block_raw,
                    timestamp = timestamp, nonce = nonce_raw, height = height,
@@ -183,7 +186,7 @@ class Block:
             'height': self.height,
             'hash': self.hash,
             'indep_hash': self.indep_hash,
-            'txs': self.txs,
+            'txs': [tx.id if type(tx) is Transaction else tx for tx in self.txs],
             'tx_root': self.tx_root,
             'wallet_list': self.wallet_list,
             'reward_addr': self.reward_addr,
@@ -234,5 +237,90 @@ class Block:
             stream.write(arbinenc(tag,                             16))
         stream.write(len(self.txs).to_bytes(2, 'big'))
         for tx in self.txs[::-1]:
-            stream.write(arbinenc(b64dec(tx),                      24))
+            if type(tx) is Transaction:
+                stream.write(tx.tobytes())
+            else:
+                stream.write(arbinenc(b64dec(tx),                  24))
         return stream.getvalue()
+
+    # ar_node_utils.erlt / validate_block
+
+    #  POW = ar_randomx_state:hash(Height, << Nonce/binary, BDS/binary >>),
+                # uses randomx_state_by_height
+                        # which calls get_state_by_height, swap_height, randomx_key
+                                    # swap height floors the height to RANDOMX_KEY_SWAP_FREQ
+                            # get_state_by_height and get_key_from_Cache also use swap_height
+                # to branch to one of
+                # ar_mine_randomx: hash_fast, hash_light, init_light;hash_light
+    #   ar_mine:validate(POW, ar_poa:modify_diff(Diff, POA#poa.option, Height), Height)
+
+    #def compute_indep_hash(self): # from ar_http_iface_middleware.erl
+    #    bds = self._get_data_segment()
+    #    return bds, ar_weave:indep_hash(BDS, self.hash_row, self.nonce_raw, self.poa)
+
+    # from ar_block.erl
+    def _get_data_segment(self):
+        bds_base = self._get_data_segment_base()
+        block_index_merkle = self.hash_list_merkle_raw
+        return deep_hash([
+            bds_base,
+            str(self.timestamp).encode(),
+            str(self.last_retarget).encode(),
+            str(self.diff).encode(),
+            str(self.cumulative_diff).encode(),
+            str(self.reward_pool).encode(),
+            self.wallet_list_raw,
+            block_index_merkle
+        ])
+    def _get_data_segment_base(self):
+        if self.height >= FORK_2_4:
+            props = [
+                str(self.height).encode(),
+                self.prev_block_raw,
+                self.tx_root_raw,
+                [
+                    b64dec(tx.id if type(tx) is Transaction else tx)
+                    for tx in self.tx
+                ],
+                str(self.block_size).encode(),
+                str(self.weave_size).encode(),
+                self.reward_addr_raw,
+                self.tags
+            ]
+            if self.height >= FORK_2_5:
+                props2 = [
+                    str(self.usd_to_ar_rate[0]).encode(),   
+                    str(self.usd_to_ar_rate[1]).encode(),   
+                    str(self.scheduled_usd_to_ar_rate[0]).encode(),
+                    str(self.scheduled_usd_to_ar_rate[1]).encode(),
+                    str(self.packing_threshold).encode(),
+                    str(self.strict_chunk_threshold),
+                    *props
+                ]
+            else:
+                props2 = props
+            return deep_hash(props2)
+        else:
+            return deep_hash([
+                str(self.height).encode(),
+                self.prev_block_raw,
+                self.tx_root_raw,
+                [
+                    b64dec(tx.id if type(tx) is Transaction else tx)
+                    for tx in self.tx
+                ],
+                str(self.block_size).encode(),
+                str(self.weave_size).encode(),
+                self.reward_addr_raw,
+                [[tag['name'].encode(), tag['value'].encode()] for tag in self.tags],
+                [
+                    str(self.poa_option).encode(),
+                    self.poa_tx_path,
+                    self.poa_data_path,
+                    self.poa_chunk
+                ]
+            ])
+    def _encode_tags(self):
+        if self.height >= FORK_2_5:
+            return 
+
